# JS Stacktraces. The good, the bad, and the ugly.

Error monitoring in Javascript is a thorny problem. On the one hand
`window.onerror` will at least notify you when something goes wrong. On the
other, it won't give you enough information to actually debug. Notably absent
in many cases is the stack trace.

## The good

This is beginning to change in Chrome and modern Opera (i.e. anything based
around Blink). In both of these browsers you can bind to the 'error' event
properly:

```javascript
// Chrome & Opera only!
window.addEventListener('error', function (e) {
   console.log(e.error.message, "from", e.error.stack);
});

```

## The bad

Unfortunately Firefox, Safari and IE are still around, so we have to support
them too. As the stacktrace is not available in `window.onerror` we have to do
a little bit more work. If we want stack-traces, we're going to have to wrap
every async entry point. This sounds impossible until you realise that all of
these browsers (IE < 8 excepted) use prototypes heavily. That means that there
are not actually that many async entry points.

```javascript
// Firefox needes Window and EventTarget, Safari / IE 11 need all of them
var targets = window.EventTarget ? [Window, EventTarget] : [Window, Node, XMLHttpRequest, ...]
targets.forEach(function (constructor) {
  ['addEventListener', 'removeEventListener'].each(function (method) {
    var original = constructor.prototype[method];
    constructor.prototype.addEventListener = function (event, callback, bubble) {
      callback.wrapped = callback.wrapped || function () {
        try{
          callback.apply(this, arguments);
        } catch(e) {
          // handle e;
          throw e;
        }
      };

      original.call(event, callback.wrapped, bubble);
    };
  };
});

// Also wrap setTimeout, setInterval, requestAnimationFrame, and a handful of others
```

## The Ugly

In IE <11, there's not much hope. Even exception objects don't have a stack
property, so even if you can rescue them that doesn't help you. If you're in IE
<8 you don't even have prototypes, (or even addEventListener!) so there's no
way such a neat hack would work.

Luckily there's a little bit we can still do. It turns out that in IE,
`window.onerror` is called as though it were a function call. This means that
you can access the stacktrace via `arguments.callee.caller.caller...`. This is
a bit of a pain, and it only gives you function names instead of line numbers,
but it's better than nothing.

```
// IE <11
window.onerror = function (error, file, line, column) {
  var column = column || (window.event && window.event.errorCharacter);
  var stack = [];
  var f = arguments.callee.caller;
  while (f) {
     stack.push(f.name);
     f = f.caller;
  }
}
```

## Install bugsnag!

We've taken some time to fully understand the options available and give you
the best of everything. Try bugsnag-js today!

